# @pytest.mark.skip(reason="AWS Suspended Account")
# # Test the following scenarios:
# # 1. Generate a Key, and use it to make a call
# # 2. Make a call with invalid key, expect it to fail
# # 3. Make a call to a key with invalid model - expect to fail
# # 4. Make a call to a key with valid model - expect to pass
# # 5. Make a call with key over budget, expect to fail
# # 6. Make a streaming chat/completions call with key over budget, expect to fail


# # function to call to generate key - async def new_user(data: NewUserRequest):
# # function to validate a request - async def user_auth(request: Request):

# import sys, os
# import traceback
# from dotenv import load_dotenv
# from fastapi import Request

# load_dotenv()
# import os, io

# # this file is to test litellm/proxy

# sys.path.insert(
#     0, os.path.abspath("../..")
# )  # Adds the parent directory to the system path
# import pytest, logging, asyncio
# import litellm, asyncio
# from litellm.proxy.proxy_server import (
#     new_user,
#     user_api_key_auth,
#     user_update,
#     generate_key_fn,
# )

# from litellm.proxy._types import NewUserRequest, DynamoDBArgs, GenerateKeyRequest
# from litellm.proxy.utils import DBClient, hash_token
# from starlette.datastructures import URL


# request_data = {
#     "model": "azure-gpt-3.5",
#     "messages": [
#         {"role": "user", "content": "this is my new test. respond in 50 lines"}
#     ],
# }


# @pytest.fixture
# def custom_db_client():
#     # Assuming DBClient is a class that needs to be instantiated
#     db_args = {
#         "ssl_verify": False,
#         "billing_mode": "PAY_PER_REQUEST",
#         "region_name": "us-west-2",
#     }
#     custom_db_client = DBClient(
#         custom_db_type="dynamo_db",
#         custom_db_args=db_args,
#     )
#     # Reset litellm.proxy.proxy_server.prisma_client to None
#     litellm.proxy.proxy_server.prisma_client = None

#     return custom_db_client


# def test_generate_and_call_with_valid_key(custom_db_client):
#     # 1. Generate a Key, and use it to make a call
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     try:

#         async def test():
#             request = NewUserRequest()
#             key = await new_user(request)
#             print(key)

#             generated_key = key.key
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#         asyncio.run(test())
#     except Exception as e:
#         pytest.fail(f"An exception occurred - {str(e)}")


# def test_call_with_invalid_key(custom_db_client):
#     # 2. Make a call with invalid key, expect it to fail
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     try:

#         async def test():
#             generated_key = "bad-key"
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"}, receive=None)
#             request._url = URL(url="/chat/completions")

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             pytest.fail(f"This should have failed!. IT's an invalid key")

#         asyncio.run(test())
#     except Exception as e:
#         print("Got Exception", e)
#         print(e.message)
#         assert "Authentication Error" in e.message
#         pass


# def test_call_with_invalid_model(custom_db_client):
#     # 3. Make a call to a key with an invalid model - expect to fail
#     from litellm._logging import verbose_proxy_logger
#     import logging

#     verbose_proxy_logger.setLevel(logging.DEBUG)
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     try:

#         async def test():
#             request = NewUserRequest(models=["mistral"])
#             key = await new_user(request)
#             print(key)

#             generated_key = key.key
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             async def return_body():
#                 return b'{"model": "gemini-pro-vision"}'

#             request.body = return_body

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             pytest.fail(f"This should have failed!. IT's an invalid model")

#         asyncio.run(test())
#     except Exception as e:
#         assert (
#             e.message
#             == "Authentication Error, API Key not allowed to access model. This token can only access models=['mistral']. Tried to access gemini-pro-vision"
#         )
#         pass


# def test_call_with_valid_model(custom_db_client):
#     # 4. Make a call to a key with a valid model - expect to pass
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     try:

#         async def test():
#             request = NewUserRequest(models=["mistral"])
#             key = await new_user(request)
#             print(key)

#             generated_key = key.key
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             async def return_body():
#                 return b'{"model": "mistral"}'

#             request.body = return_body

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#         asyncio.run(test())
#     except Exception as e:
#         pytest.fail(f"An exception occurred - {str(e)}")


# def test_call_with_user_over_budget(custom_db_client):
#     # 5. Make a call with a key over budget, expect to fail
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     from litellm._logging import verbose_proxy_logger, verbose_logger
#     import logging

#     litellm.set_verbose = True
#     verbose_logger.setLevel(logging.DEBUG)
#     verbose_proxy_logger.setLevel(logging.DEBUG)
#     try:

#         async def test():
#             request = NewUserRequest(max_budget=0.00001)
#             key = await new_user(request)
#             print(key)

#             generated_key = key.key
#             user_id = key.user_id
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#             # update spend using track_cost callback, make 2nd request, it should fail
#             from litellm.proxy.proxy_server import (
#                 _PROXY_track_cost_callback as track_cost_callback,
#             )
#             from litellm import ModelResponse, Choices, Message, Usage

#             resp = ModelResponse(
#                 id="chatcmpl-e41836bb-bb8b-4df2-8e70-8f3e160155ac",
#                 choices=[
#                     Choices(
#                         finish_reason=None,
#                         index=0,
#                         message=Message(
#                             content=" Sure! Here is a short poem about the sky:\n\nA canvas of blue, a",
#                             role="assistant",
#                         ),
#                     )
#                 ],
#                 model="gpt-35-turbo",  # azure always has model written like this
#                 usage=Usage(prompt_tokens=210, completion_tokens=200, total_tokens=410),
#             )
#             await track_cost_callback(
#                 kwargs={
#                     "stream": False,
#                     "litellm_params": {
#                         "metadata": {
#                             "user_api_key": hash_token(generated_key),
#                             "user_api_key_user_id": user_id,
#                         }
#                     },
#                     "response_cost": 0.00002,
#                 },
#                 completion_response=resp,
#             )
#             await asyncio.sleep(5)
#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)
#             pytest.fail(f"This should have failed!. They key crossed it's budget")

#         asyncio.run(test())
#     except Exception as e:
#         error_detail = e.message
#         assert "Authentication Error, ExceededBudget:" in error_detail
#         print(vars(e))


# def test_call_with_user_over_budget_stream(custom_db_client):
#     # 6. Make a call with a key over budget, expect to fail
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     from litellm._logging import verbose_proxy_logger
#     import logging

#     litellm.set_verbose = True
#     verbose_proxy_logger.setLevel(logging.DEBUG)
#     try:

#         async def test():
#             request = NewUserRequest(max_budget=0.00001)
#             key = await new_user(request)
#             print(key)

#             generated_key = key.key
#             user_id = key.user_id
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#             # update spend using track_cost callback, make 2nd request, it should fail
#             from litellm.proxy.proxy_server import (
#                 _PROXY_track_cost_callback as track_cost_callback,
#             )
#             from litellm import ModelResponse, Choices, Message, Usage

#             resp = ModelResponse(
#                 id="chatcmpl-e41836bb-bb8b-4df2-8e70-8f3e160155ac",
#                 choices=[
#                     Choices(
#                         finish_reason=None,
#                         index=0,
#                         message=Message(
#                             content=" Sure! Here is a short poem about the sky:\n\nA canvas of blue, a",
#                             role="assistant",
#                         ),
#                     )
#                 ],
#                 model="gpt-35-turbo",  # azure always has model written like this
#                 usage=Usage(prompt_tokens=210, completion_tokens=200, total_tokens=410),
#             )
#             await track_cost_callback(
#                 kwargs={
#                     "stream": True,
#                     "complete_streaming_response": resp,
#                     "litellm_params": {
#                         "metadata": {
#                             "user_api_key": hash_token(generated_key),
#                             "user_api_key_user_id": user_id,
#                         }
#                     },
#                     "response_cost": 0.00002,
#                 },
#                 completion_response=ModelResponse(),
#             )
#             await asyncio.sleep(5)
#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)
#             pytest.fail(f"This should have failed!. They key crossed it's budget")

#         asyncio.run(test())
#     except Exception as e:
#         error_detail = e.message
#         assert "Authentication Error, ExceededBudget:" in error_detail
#         print(vars(e))


# def test_call_with_user_key_budget(custom_db_client):
#     # 7. Make a call with a key over budget, expect to fail
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     from litellm._logging import verbose_proxy_logger
#     import logging

#     verbose_proxy_logger.setLevel(logging.DEBUG)
#     try:

#         async def test():
#             request = GenerateKeyRequest(max_budget=0.00001)
#             key = await generate_key_fn(request)
#             print(key)

#             generated_key = key.key
#             user_id = key.user_id
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#             # update spend using track_cost callback, make 2nd request, it should fail
#             from litellm.proxy.proxy_server import (
#                 _PROXY_track_cost_callback as track_cost_callback,
#             )
#             from litellm import ModelResponse, Choices, Message, Usage

#             resp = ModelResponse(
#                 id="chatcmpl-e41836bb-bb8b-4df2-8e70-8f3e160155ac",
#                 choices=[
#                     Choices(
#                         finish_reason=None,
#                         index=0,
#                         message=Message(
#                             content=" Sure! Here is a short poem about the sky:\n\nA canvas of blue, a",
#                             role="assistant",
#                         ),
#                     )
#                 ],
#                 model="gpt-35-turbo",  # azure always has model written like this
#                 usage=Usage(prompt_tokens=210, completion_tokens=200, total_tokens=410),
#             )
#             await track_cost_callback(
#                 kwargs={
#                     "stream": False,
#                     "litellm_params": {
#                         "metadata": {
#                             "user_api_key": hash_token(generated_key),
#                             "user_api_key_user_id": user_id,
#                         }
#                     },
#                     "response_cost": 0.00002,
#                 },
#                 completion_response=resp,
#             )
#             await asyncio.sleep(5)
#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)
#             pytest.fail(f"This should have failed!. They key crossed it's budget")

#         asyncio.run(test())
#     except Exception as e:
#         error_detail = e.message
#         assert "Authentication Error, ExceededTokenBudget:" in error_detail
#         print(vars(e))


# def test_call_with_key_over_budget_stream(custom_db_client):
#     # 8. Make a call with a key over budget, expect to fail
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     from litellm._logging import verbose_proxy_logger
#     import logging

#     litellm.set_verbose = True
#     verbose_proxy_logger.setLevel(logging.DEBUG)
#     try:

#         async def test():
#             request = GenerateKeyRequest(max_budget=0.00001)
#             key = await generate_key_fn(request)
#             print(key)

#             generated_key = key.key
#             user_id = key.user_id
#             bearer_token = "Bearer " + generated_key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#             # update spend using track_cost callback, make 2nd request, it should fail
#             from litellm.proxy.proxy_server import (
#                 _PROXY_track_cost_callback as track_cost_callback,
#             )
#             from litellm import ModelResponse, Choices, Message, Usage

#             resp = ModelResponse(
#                 id="chatcmpl-e41836bb-bb8b-4df2-8e70-8f3e160155ac",
#                 choices=[
#                     Choices(
#                         finish_reason=None,
#                         index=0,
#                         message=Message(
#                             content=" Sure! Here is a short poem about the sky:\n\nA canvas of blue, a",
#                             role="assistant",
#                         ),
#                     )
#                 ],
#                 model="gpt-35-turbo",  # azure always has model written like this
#                 usage=Usage(prompt_tokens=210, completion_tokens=200, total_tokens=410),
#             )
#             await track_cost_callback(
#                 kwargs={
#                     "stream": True,
#                     "complete_streaming_response": resp,
#                     "litellm_params": {
#                         "metadata": {
#                             "user_api_key": hash_token(generated_key),
#                             "user_api_key_user_id": user_id,
#                         }
#                     },
#                     "response_cost": 0.00002,
#                 },
#                 completion_response=ModelResponse(),
#             )
#             await asyncio.sleep(5)
#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)
#             pytest.fail(f"This should have failed!. They key crossed it's budget")

#         asyncio.run(test())
#     except Exception as e:
#         error_detail = e.message
#         assert "Authentication Error, ExceededTokenBudget:" in error_detail
#         print(vars(e))


# def test_dynamo_db_migration(custom_db_client):
#     # Tests the temporary patch we have in place
#     setattr(litellm.proxy.proxy_server, "custom_db_client", custom_db_client)
#     setattr(litellm.proxy.proxy_server, "master_key", "sk-1234")
#     setattr(litellm.proxy.proxy_server, "user_custom_auth", None)
#     try:

#         async def test():
#             request = GenerateKeyRequest(max_budget=1)
#             key = await generate_key_fn(request)
#             print(key)

#             generated_key = key.key
#             bearer_token = (
#                 "Bearer " + generated_key
#             )  # this works with ishaan's db, it's a never expiring key

#             request = Request(scope={"type": "http"})
#             request._url = URL(url="/chat/completions")

#             async def return_body():
#                 return b'{"model": "azure-models"}'

#             request.body = return_body

#             # use generated key to auth in
#             result = await user_api_key_auth(request=request, api_key=bearer_token)
#             print("result from user auth with new key", result)

#         asyncio.run(test())
#     except Exception as e:
#         pytest.fail(f"An exception occurred - {traceback.format_exc()}")



from typing import Optional
import httpx


class HTTPHandler:
    def __init__(self, concurrent_limit=1000):
        # Create a client with a connection pool
        self.client = httpx.AsyncClient(
            limits=httpx.Limits(
                max_connections=concurrent_limit,
                max_keepalive_connections=concurrent_limit,
            )
        )

    async def close(self):
        # Close the client when you're done with it
        await self.client.aclose()

    async def get(
        self, url: str, params: Optional[dict] = None, headers: Optional[dict] = None
    ):
        response = await self.client.get(url, params=params, headers=headers)
        return response

    async def post(
        self,
        url: str,
        data: Optional[dict] = None,
        params: Optional[dict] = None,
        headers: Optional[dict] = None,
    ):
        try:
            response = await self.client.post(
                url, data=data, params=params, headers=headers
            )
            return response
        except Exception as e:
            raise e



# What is this?
## Checks TPM/RPM Limits for a key/user/team on the proxy
## Works with Redis - if given

from typing import Optional, Literal
import litellm, traceback, sys
from litellm.caching import DualCache, RedisCache
from litellm.proxy._types import (
    UserAPIKeyAuth,
    LiteLLM_VerificationTokenView,
    LiteLLM_UserTable,
    LiteLLM_TeamTable,
)
from litellm.integrations.custom_logger import CustomLogger
from fastapi import HTTPException
from litellm._logging import verbose_proxy_logger
from litellm import ModelResponse
from datetime import datetime


class _PROXY_MaxTPMRPMLimiter(CustomLogger):
    user_api_key_cache = None

    # Class variables or attributes
    def __init__(self, internal_cache: Optional[DualCache]):
        if internal_cache is None:
            self.internal_cache = DualCache()
        else:
            self.internal_cache = internal_cache

    def print_verbose(self, print_statement):
        try:
            verbose_proxy_logger.debug(print_statement)
            if litellm.set_verbose:
                print(print_statement)  # noqa
        except:
            pass

            ## check if admin has set tpm/rpm limits for this key/user/team

    def _check_limits_set(
        self,
        user_api_key_cache: DualCache,
        key: Optional[str],
        user_id: Optional[str],
        team_id: Optional[str],
    ) -> bool:
        ## key
        if key is not None:
            key_val = user_api_key_cache.get_cache(key=key)
            if isinstance(key_val, dict):
                key_val = LiteLLM_VerificationTokenView(**key_val)

            if isinstance(key_val, LiteLLM_VerificationTokenView):
                user_api_key_tpm_limit = key_val.tpm_limit

                user_api_key_rpm_limit = key_val.rpm_limit

                if (
                    user_api_key_tpm_limit is not None
                    or user_api_key_rpm_limit is not None
                ):
                    return True

        ## team
        if team_id is not None:
            team_val = user_api_key_cache.get_cache(key=team_id)
            if isinstance(team_val, dict):
                team_val = LiteLLM_TeamTable(**team_val)

            if isinstance(team_val, LiteLLM_TeamTable):
                team_tpm_limit = team_val.tpm_limit

                team_rpm_limit = team_val.rpm_limit

                if team_tpm_limit is not None or team_rpm_limit is not None:
                    return True

        ## user
        if user_id is not None:
            user_val = user_api_key_cache.get_cache(key=user_id)
            if isinstance(user_val, dict):
                user_val = LiteLLM_UserTable(**user_val)

            if isinstance(user_val, LiteLLM_UserTable):
                user_tpm_limit = user_val.tpm_limit

                user_rpm_limit = user_val.rpm_limit

                if user_tpm_limit is not None or user_rpm_limit is not None:
                    return True
        return False

    async def check_key_in_limits(
        self,
        user_api_key_dict: UserAPIKeyAuth,
        current_minute_dict: dict,
        tpm_limit: int,
        rpm_limit: int,
        request_count_api_key: str,
        type: Literal["key", "user", "team"],
    ):

        if type == "key" and user_api_key_dict.api_key is not None:
            current = current_minute_dict["key"].get(user_api_key_dict.api_key, None)
        elif type == "user" and user_api_key_dict.user_id is not None:
            current = current_minute_dict["user"].get(user_api_key_dict.user_id, None)
        elif type == "team" and user_api_key_dict.team_id is not None:
            current = current_minute_dict["team"].get(user_api_key_dict.team_id, None)
        else:
            return
        if current is None:
            if tpm_limit == 0 or rpm_limit == 0:
                # base case
                raise HTTPException(
                    status_code=429, detail="Max tpm/rpm limit reached."
                )
        elif current["current_tpm"] < tpm_limit and current["current_rpm"] < rpm_limit:
            pass
        else:
            raise HTTPException(status_code=429, detail="Max tpm/rpm limit reached.")

    async def async_pre_call_hook(
        self,
        user_api_key_dict: UserAPIKeyAuth,
        cache: DualCache,
        data: dict,
        call_type: str,
    ):
        self.print_verbose(
            f"Inside Max TPM/RPM Limiter Pre-Call Hook - {user_api_key_dict}"
        )
        api_key = user_api_key_dict.api_key
        # check if REQUEST ALLOWED for user_id
        user_id = user_api_key_dict.user_id
        ## get team tpm/rpm limits
        team_id = user_api_key_dict.team_id

        self.user_api_key_cache = cache

        _set_limits = self._check_limits_set(
            user_api_key_cache=cache, key=api_key, user_id=user_id, team_id=team_id
        )

        self.print_verbose(f"_set_limits: {_set_limits}")

        if _set_limits == False:
            return

        # ------------
        # Setup values
        # ------------

        current_date = datetime.now().strftime("%Y-%m-%d")
        current_hour = datetime.now().strftime("%H")
        current_minute = datetime.now().strftime("%M")
        precise_minute = f"{current_date}-{current_hour}-{current_minute}"
        cache_key = "usage:{}".format(precise_minute)
        current_minute_dict = await self.internal_cache.async_get_cache(
            key=cache_key
        )  # {"usage:{curr_minute}": {"key": {<api_key>: {"current_requests": 1, "current_tpm": 1, "current_rpm": 10}}}}

        if current_minute_dict is None:
            current_minute_dict = {"key": {}, "user": {}, "team": {}}

        if api_key is not None:
            tpm_limit = getattr(user_api_key_dict, "tpm_limit", sys.maxsize)
            if tpm_limit is None:
                tpm_limit = sys.maxsize
            rpm_limit = getattr(user_api_key_dict, "rpm_limit", sys.maxsize)
            if rpm_limit is None:
                rpm_limit = sys.maxsize
            request_count_api_key = f"{api_key}::{precise_minute}::request_count"
            await self.check_key_in_limits(
                user_api_key_dict=user_api_key_dict,
                current_minute_dict=current_minute_dict,
                request_count_api_key=request_count_api_key,
                tpm_limit=tpm_limit,
                rpm_limit=rpm_limit,
                type="key",
            )

        if user_id is not None:
            _user_id_rate_limits = user_api_key_dict.user_id_rate_limits

            # get user tpm/rpm limits
            if _user_id_rate_limits is not None and isinstance(
                _user_id_rate_limits, dict
            ):
                user_tpm_limit = _user_id_rate_limits.get("tpm_limit", None)
                user_rpm_limit = _user_id_rate_limits.get("rpm_limit", None)
                if user_tpm_limit is None:
                    user_tpm_limit = sys.maxsize
                if user_rpm_limit is None:
                    user_rpm_limit = sys.maxsize

                # now do the same tpm/rpm checks
                request_count_api_key = f"{user_id}::{precise_minute}::request_count"
                # print(f"Checking if {request_count_api_key} is allowed to make request for minute {precise_minute}")
                await self.check_key_in_limits(
                    user_api_key_dict=user_api_key_dict,
                    current_minute_dict=current_minute_dict,
                    request_count_api_key=request_count_api_key,
                    tpm_limit=user_tpm_limit,
                    rpm_limit=user_rpm_limit,
                    type="user",
                )

        # TEAM RATE LIMITS
        if team_id is not None:
            team_tpm_limit = getattr(user_api_key_dict, "team_tpm_limit", sys.maxsize)

            if team_tpm_limit is None:
                team_tpm_limit = sys.maxsize
            team_rpm_limit = getattr(user_api_key_dict, "team_rpm_limit", sys.maxsize)
            if team_rpm_limit is None:
                team_rpm_limit = sys.maxsize

            if team_tpm_limit is None:
                team_tpm_limit = sys.maxsize
            if team_rpm_limit is None:
                team_rpm_limit = sys.maxsize

            # now do the same tpm/rpm checks
            request_count_api_key = f"{team_id}::{precise_minute}::request_count"

            # print(f"Checking if {request_count_api_key} is allowed to make request for minute {precise_minute}")
            await self.check_key_in_limits(
                user_api_key_dict=user_api_key_dict,
                current_minute_dict=current_minute_dict,
                request_count_api_key=request_count_api_key,
                tpm_limit=team_tpm_limit,
                rpm_limit=team_rpm_limit,
                type="team",
            )

        return

    async def async_log_success_event(self, kwargs, response_obj, start_time, end_time):
        try:
            self.print_verbose(f"INSIDE TPM RPM Limiter ASYNC SUCCESS LOGGING")

            user_api_key = kwargs["litellm_params"]["metadata"]["user_api_key"]
            user_api_key_user_id = kwargs["litellm_params"]["metadata"].get(
                "user_api_key_user_id", None
            )
            user_api_key_team_id = kwargs["litellm_params"]["metadata"].get(
                "user_api_key_team_id", None
            )
            _limits_set = self._check_limits_set(
                user_api_key_cache=self.user_api_key_cache,
                key=user_api_key,
                user_id=user_api_key_user_id,
                team_id=user_api_key_team_id,
            )

            if _limits_set == False:  # don't waste cache calls if no tpm/rpm limits set
                return

            # ------------
            # Setup values
            # ------------

            current_date = datetime.now().strftime("%Y-%m-%d")
            current_hour = datetime.now().strftime("%H")
            current_minute = datetime.now().strftime("%M")
            precise_minute = f"{current_date}-{current_hour}-{current_minute}"

            total_tokens = 0

            if isinstance(response_obj, ModelResponse):
                total_tokens = response_obj.usage.total_tokens

            """
            - get value from redis
            - increment requests + 1
            - increment tpm + 1
            - increment rpm + 1
            - update value in-memory + redis
            """
            cache_key = "usage:{}".format(precise_minute)
            if (
                self.internal_cache.redis_cache is not None
            ):  # get straight from redis if possible
                current_minute_dict = (
                    await self.internal_cache.redis_cache.async_get_cache(
                        key=cache_key,
                    )
                )  # {"usage:{current_minute}": {"key": {}, "team": {}, "user": {}}}
            else:
                current_minute_dict = await self.internal_cache.async_get_cache(
                    key=cache_key,
                )

            if current_minute_dict is None:
                current_minute_dict = {"key": {}, "user": {}, "team": {}}

            _cache_updated = False  # check if a cache update is required. prevent unnecessary rewrites.

            # ------------
            # Update usage - API Key
            # ------------

            if user_api_key is not None:
                _cache_updated = True
                ## API KEY ##
                if user_api_key in current_minute_dict["key"]:
                    current_key_usage = current_minute_dict["key"][user_api_key]
                    new_val = {
                        "current_tpm": current_key_usage["current_tpm"] + total_tokens,
                        "current_rpm": current_key_usage["current_rpm"] + 1,
                    }
                else:
                    new_val = {
                        "current_tpm": total_tokens,
                        "current_rpm": 1,
                    }

                current_minute_dict["key"][user_api_key] = new_val

                self.print_verbose(
                    f"updated_value in success call: {new_val}, precise_minute: {precise_minute}"
                )

            # ------------
            # Update usage - User
            # ------------
            if user_api_key_user_id is not None:
                _cache_updated = True
                total_tokens = 0

                if isinstance(response_obj, ModelResponse):
                    total_tokens = response_obj.usage.total_tokens

                if user_api_key_user_id in current_minute_dict["key"]:
                    current_key_usage = current_minute_dict["key"][user_api_key_user_id]
                    new_val = {
                        "current_tpm": current_key_usage["current_tpm"] + total_tokens,
                        "current_rpm": current_key_usage["current_rpm"] + 1,
                    }
                else:
                    new_val = {
                        "current_tpm": total_tokens,
                        "current_rpm": 1,
                    }

                current_minute_dict["user"][user_api_key_user_id] = new_val

            # ------------
            # Update usage - Team
            # ------------
            if user_api_key_team_id is not None:
                _cache_updated = True
                total_tokens = 0

                if isinstance(response_obj, ModelResponse):
                    total_tokens = response_obj.usage.total_tokens

                if user_api_key_team_id in current_minute_dict["key"]:
                    current_key_usage = current_minute_dict["key"][user_api_key_team_id]
                    new_val = {
                        "current_tpm": current_key_usage["current_tpm"] + total_tokens,
                        "current_rpm": current_key_usage["current_rpm"] + 1,
                    }
                else:
                    new_val = {
                        "current_tpm": total_tokens,
                        "current_rpm": 1,
                    }

                current_minute_dict["team"][user_api_key_team_id] = new_val

            if _cache_updated == True:
                await self.internal_cache.async_set_cache(
                    key=cache_key, value=current_minute_dict
                )

        except Exception as e:
            self.print_verbose("{}\n{}".format(e, traceback.format_exc()))  # noqa



# What is this?
## Common auth checks between jwt + key based auth
"""
Got Valid Token from Cache, DB
Run checks for: 

1. If user can call model
2. If user is in budget 
3. If end_user ('user' passed to /chat/completions, /embeddings endpoint) is in budget 
"""
from litellm.proxy._types import (
    LiteLLM_UserTable,
    LiteLLM_EndUserTable,
    LiteLLM_JWTAuth,
    LiteLLM_TeamTable,
    LiteLLMRoutes,
)
from typing import Optional, Literal, Union
from litellm.proxy.utils import PrismaClient
from litellm.caching import DualCache
import litellm

all_routes = LiteLLMRoutes.openai_routes.value + LiteLLMRoutes.management_routes.value


def common_checks(
    request_body: dict,
    team_object: LiteLLM_TeamTable,
    end_user_object: Optional[LiteLLM_EndUserTable],
    global_proxy_spend: Optional[float],
    general_settings: dict,
    route: str,
) -> bool:
    """
    Common checks across jwt + key-based auth.

    1. If team is blocked
    2. If team can call model
    3. If team is in budget
    4. If end_user ('user' passed to /chat/completions, /embeddings endpoint) is in budget
    5. [OPTIONAL] If 'enforce_end_user' enabled - did developer pass in 'user' param for openai endpoints
    6. [OPTIONAL] If 'litellm.max_budget' is set (>0), is proxy under budget
    """
    _model = request_body.get("model", None)
    if team_object.blocked == True:
        raise Exception(
            f"Team={team_object.team_id} is blocked. Update via `/team/unblock` if your admin."
        )
    # 2. If user can call model
    if (
        _model is not None
        and len(team_object.models) > 0
        and _model not in team_object.models
    ):
        # this means the team has access to all models on the proxy
        if "all-proxy-models" in team_object.models:
            # this means the team has access to all models on the proxy
            pass
        else:
            raise Exception(
                f"Team={team_object.team_id} not allowed to call model={_model}. Allowed team models = {team_object.models}"
            )
    # 3. If team is in budget
    if (
        team_object.max_budget is not None
        and team_object.spend is not None
        and team_object.spend > team_object.max_budget
    ):
        raise Exception(
            f"Team={team_object.team_id} over budget. Spend={team_object.spend}, Budget={team_object.max_budget}"
        )
    # 4. If end_user ('user' passed to /chat/completions, /embeddings endpoint) is in budget
    if end_user_object is not None and end_user_object.litellm_budget_table is not None:
        end_user_budget = end_user_object.litellm_budget_table.max_budget
        if end_user_budget is not None and end_user_object.spend > end_user_budget:
            raise Exception(
                f"ExceededBudget: End User={end_user_object.user_id} over budget. Spend={end_user_object.spend}, Budget={end_user_budget}"
            )
    # 5. [OPTIONAL] If 'enforce_user_param' enabled - did developer pass in 'user' param for openai endpoints
    if (
        general_settings.get("enforce_user_param", None) is not None
        and general_settings["enforce_user_param"] == True
    ):
        if route in LiteLLMRoutes.openai_routes.value and "user" not in request_body:
            raise Exception(
                f"'user' param not passed in. 'enforce_user_param'={general_settings['enforce_user_param']}"
            )
    # 6. [OPTIONAL] If 'litellm.max_budget' is set (>0), is proxy under budget
    if litellm.max_budget > 0 and global_proxy_spend is not None:
        if global_proxy_spend > litellm.max_budget:
            raise Exception(
                f"ExceededBudget: LiteLLM Proxy has exceeded its budget. Current spend: {global_proxy_spend}; Max Budget: {litellm.max_budget}"
            )
    return True


def _allowed_routes_check(user_route: str, allowed_routes: list) -> bool:
    for allowed_route in allowed_routes:
        if (
            allowed_route == LiteLLMRoutes.openai_routes.name
            and user_route in LiteLLMRoutes.openai_routes.value
        ):
            return True
        elif (
            allowed_route == LiteLLMRoutes.info_routes.name
            and user_route in LiteLLMRoutes.info_routes.value
        ):
            return True
        elif (
            allowed_route == LiteLLMRoutes.management_routes.name
            and user_route in LiteLLMRoutes.management_routes.value
        ):
            return True
        elif allowed_route == user_route:
            return True
    return False


def allowed_routes_check(
    user_role: Literal["proxy_admin", "team"],
    user_route: str,
    litellm_proxy_roles: LiteLLM_JWTAuth,
) -> bool:
    """
    Check if user -> not admin - allowed to access these routes
    """

    if user_role == "proxy_admin":
        if litellm_proxy_roles.admin_allowed_routes is None:
            is_allowed = _allowed_routes_check(
                user_route=user_route, allowed_routes=["management_routes"]
            )
            return is_allowed
        elif litellm_proxy_roles.admin_allowed_routes is not None:
            is_allowed = _allowed_routes_check(
                user_route=user_route,
                allowed_routes=litellm_proxy_roles.admin_allowed_routes,
            )
            return is_allowed

    elif user_role == "team":
        if litellm_proxy_roles.team_allowed_routes is None:
            """
            By default allow a team to call openai + info routes
            """
            is_allowed = _allowed_routes_check(
                user_route=user_route, allowed_routes=["openai_routes", "info_routes"]
            )
            return is_allowed
        elif litellm_proxy_roles.team_allowed_routes is not None:
            is_allowed = _allowed_routes_check(
                user_route=user_route,
                allowed_routes=litellm_proxy_roles.team_allowed_routes,
            )
            return is_allowed
    return False


def get_actual_routes(allowed_routes: list) -> list:
    actual_routes: list = []
    for route_name in allowed_routes:
        try:
            route_value = LiteLLMRoutes[route_name].value
            actual_routes = actual_routes + route_value
        except KeyError:
            actual_routes.append(route_name)
    return actual_routes


async def get_end_user_object(
    end_user_id: Optional[str],
    prisma_client: Optional[PrismaClient],
    user_api_key_cache: DualCache,
) -> Optional[LiteLLM_EndUserTable]:
    """
    Returns end user object, if in db.

    Do a isolated check for end user in table vs. doing a combined key + team + user + end-user check, as key might come in frequently for different end-users. Larger call will slowdown query time. This way we get to cache the constant (key/team/user info) and only update based on the changing value (end-user).
    """
    if prisma_client is None:
        raise Exception("No db connected")

    if end_user_id is None:
        return None

    # check if in cache
    cached_user_obj = user_api_key_cache.async_get_cache(key=end_user_id)
    if cached_user_obj is not None:
        if isinstance(cached_user_obj, dict):
            return LiteLLM_EndUserTable(**cached_user_obj)
        elif isinstance(cached_user_obj, LiteLLM_EndUserTable):
            return cached_user_obj
    # else, check db
    try:
        response = await prisma_client.db.litellm_endusertable.find_unique(
            where={"user_id": end_user_id}
        )

        if response is None:
            raise Exception

        return LiteLLM_EndUserTable(**response.dict())
    except Exception as e:  # if end-user not in db
        return None


async def get_user_object(self, user_id: str) -> LiteLLM_UserTable:
    """
    - Check if user id in proxy User Table
    - if valid, return LiteLLM_UserTable object with defined limits
    - if not, then raise an error
    """
    if self.prisma_client is None:
        raise Exception(
            "No DB Connected. See - https://docs.litellm.ai/docs/proxy/virtual_keys"
        )

    # check if in cache
    cached_user_obj = self.user_api_key_cache.async_get_cache(key=user_id)
    if cached_user_obj is not None:
        if isinstance(cached_user_obj, dict):
            return LiteLLM_UserTable(**cached_user_obj)
        elif isinstance(cached_user_obj, LiteLLM_UserTable):
            return cached_user_obj
    # else, check db
    try:
        response = await self.prisma_client.db.litellm_usertable.find_unique(
            where={"user_id": user_id}
        )

        if response is None:
            raise Exception

        return LiteLLM_UserTable(**response.dict())
    except Exception as e:
        raise Exception(
            f"User doesn't exist in db. User={user_id}. Create user via `/user/new` call."
        )


async def get_team_object(
    team_id: str,
    prisma_client: Optional[PrismaClient],
    user_api_key_cache: DualCache,
) -> LiteLLM_TeamTable:
    """
    - Check if team id in proxy Team Table
    - if valid, return LiteLLM_TeamTable object with defined limits
    - if not, then raise an error
    """
    if prisma_client is None:
        raise Exception(
            "No DB Connected. See - https://docs.litellm.ai/docs/proxy/virtual_keys"
        )

    # check if in cache
    cached_team_obj = user_api_key_cache.async_get_cache(key=team_id)
    if cached_team_obj is not None:
        if isinstance(cached_team_obj, dict):
            return LiteLLM_TeamTable(**cached_team_obj)
        elif isinstance(cached_team_obj, LiteLLM_TeamTable):
            return cached_team_obj
    # else, check db
    try:
        response = await prisma_client.db.litellm_teamtable.find_unique(
            where={"team_id": team_id}
        )

        if response is None:
            raise Exception

        return LiteLLM_TeamTable(**response.dict())
    except Exception as e:
        raise Exception(
            f"Team doesn't exist in db. Team={team_id}. Create team via `/team/new` call."
        )



#### What this tests ####
#    This tests if ahealth_check() actually works

import sys, os
import traceback
import pytest

sys.path.insert(
    0, os.path.abspath("../..")
)  # Adds the parent directory to the system path
import litellm, asyncio


@pytest.mark.asyncio
async def test_azure_health_check():
    response = await litellm.ahealth_check(
        model_params={
            "model": "azure/chatgpt-v-2",
            "messages": [{"role": "user", "content": "Hey, how's it going?"}],
            "api_key": os.getenv("AZURE_API_KEY"),
            "api_base": os.getenv("AZURE_API_BASE"),
            "api_version": os.getenv("AZURE_API_VERSION"),
        }
    )
    print(f"response: {response}")

    assert "x-ratelimit-remaining-tokens" in response
    return response


# asyncio.run(test_azure_health_check())


@pytest.mark.asyncio
async def test_azure_embedding_health_check():
    response = await litellm.ahealth_check(
        model_params={
            "model": "azure/azure-embedding-model",
            "api_key": os.getenv("AZURE_API_KEY"),
            "api_base": os.getenv("AZURE_API_BASE"),
            "api_version": os.getenv("AZURE_API_VERSION"),
        },
        input=["test for litellm"],
        mode="embedding",
    )
    print(f"response: {response}")

    assert "x-ratelimit-remaining-tokens" in response
    return response


@pytest.mark.asyncio
async def test_openai_img_gen_health_check():
    response = await litellm.ahealth_check(
        model_params={
            "model": "dall-e-3",
            "api_key": os.getenv("OPENAI_API_KEY"),
        },
        mode="image_generation",
        prompt="cute baby sea otter",
    )
    print(f"response: {response}")

    assert isinstance(response, dict) and "error" not in response
    return response


# asyncio.run(test_openai_img_gen_health_check())


async def test_azure_img_gen_health_check():
    response = await litellm.ahealth_check(
        model_params={
            "model": "azure/",
            "api_base": os.getenv("AZURE_API_BASE"),
            "api_key": os.getenv("AZURE_API_KEY"),
            "api_version": "2023-06-01-preview",
        },
        mode="image_generation",
        prompt="cute baby sea otter",
    )

    assert isinstance(response, dict) and "error" not in response
    return response


# asyncio.run(test_azure_img_gen_health_check())


@pytest.mark.skip(reason="AWS Suspended Account")
@pytest.mark.asyncio
async def test_sagemaker_embedding_health_check():
    response = await litellm.ahealth_check(
        model_params={
            "model": "sagemaker/berri-benchmarking-gpt-j-6b-fp16",
            "messages": [{"role": "user", "content": "Hey, how's it going?"}],
        },
        mode="embedding",
        input=["test from litellm"],
    )
    print(f"response: {response}")

    assert isinstance(response, dict)
    return response


# asyncio.run(test_sagemaker_embedding_health_check())



import os, types
import json
from enum import Enum
import requests
import time, traceback
from typing import Callable, Optional
from litellm.utils import ModelResponse, Choices, Message, Usage
import litellm
import httpx
from .prompt_templates.factory import cohere_message_pt


class CohereError(Exception):
    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message
        self.request = httpx.Request(method="POST", url="https://api.cohere.ai/v1/chat")
        self.response = httpx.Response(status_code=status_code, request=self.request)
        super().__init__(
            self.message
        )  # Call the base class constructor with the parameters it needs


class CohereChatConfig:
    """
    Configuration class for Cohere's API interface.

    Args:
        preamble (str, optional): When specified, the default Cohere preamble will be replaced with the provided one.
        chat_history (List[Dict[str, str]], optional): A list of previous messages between the user and the model.
        generation_id (str, optional): Unique identifier for the generated reply.
        response_id (str, optional): Unique identifier for the response.
        conversation_id (str, optional): An alternative to chat_history, creates or resumes a persisted conversation.
        prompt_truncation (str, optional): Dictates how the prompt will be constructed. Options: 'AUTO', 'AUTO_PRESERVE_ORDER', 'OFF'.
        connectors (List[Dict[str, str]], optional): List of connectors (e.g., web-search) to enrich the model's reply.
        search_queries_only (bool, optional): When true, the response will only contain a list of generated search queries.
        documents (List[Dict[str, str]], optional): A list of relevant documents that the model can cite.
        temperature (float, optional): A non-negative float that tunes the degree of randomness in generation.
        max_tokens (int, optional): The maximum number of tokens the model will generate as part of the response.
        k (int, optional): Ensures only the top k most likely tokens are considered for generation at each step.
        p (float, optional): Ensures that only the most likely tokens, with total probability mass of p, are considered for generation.
        frequency_penalty (float, optional): Used to reduce repetitiveness of generated tokens.
        presence_penalty (float, optional): Used to reduce repetitiveness of generated tokens.
        tools (List[Dict[str, str]], optional): A list of available tools (functions) that the model may suggest invoking.
        tool_results (List[Dict[str, Any]], optional): A list of results from invoking tools.
    """

    preamble: Optional[str] = None
    chat_history: Optional[list] = None
    generation_id: Optional[str] = None
    response_id: Optional[str] = None
    conversation_id: Optional[str] = None
    prompt_truncation: Optional[str] = None
    connectors: Optional[list] = None
    search_queries_only: Optional[bool] = None
    documents: Optional[list] = None
    temperature: Optional[int] = None
    max_tokens: Optional[int] = None
    k: Optional[int] = None
    p: Optional[int] = None
    frequency_penalty: Optional[int] = None
    presence_penalty: Optional[int] = None
    tools: Optional[list] = None
    tool_results: Optional[list] = None

    def __init__(
        self,
        preamble: Optional[str] = None,
        chat_history: Optional[list] = None,
        generation_id: Optional[str] = None,
        response_id: Optional[str] = None,
        conversation_id: Optional[str] = None,
        prompt_truncation: Optional[str] = None,
        connectors: Optional[list] = None,
        search_queries_only: Optional[bool] = None,
        documents: Optional[list] = None,
        temperature: Optional[int] = None,
        max_tokens: Optional[int] = None,
        k: Optional[int] = None,
        p: Optional[int] = None,
        frequency_penalty: Optional[int] = None,
        presence_penalty: Optional[int] = None,
        tools: Optional[list] = None,
        tool_results: Optional[list] = None,
    ) -> None:
        locals_ = locals()
        for key, value in locals_.items():
            if key != "self" and value is not None:
                setattr(self.__class__, key, value)

    @classmethod
    def get_config(cls):
        return {
            k: v
            for k, v in cls.__dict__.items()
            if not k.startswith("__")
            and not isinstance(
                v,
                (
                    types.FunctionType,
                    types.BuiltinFunctionType,
                    classmethod,
                    staticmethod,
                ),
            )
            and v is not None
        }


def validate_environment(api_key):
    headers = {
        "accept": "application/json",
        "content-type": "application/json",
    }
    if api_key:
        headers["Authorization"] = f"Bearer {api_key}"
    return headers


def translate_openai_tool_to_cohere(openai_tool):
    # cohere tools look like this
    """
    {
       "name": "query_daily_sales_report",
       "description": "Connects to a database to retrieve overall sales volumes and sales information for a given day.",
       "parameter_definitions": {
           "day": {
               "description": "Retrieves sales data for this day, formatted as YYYY-MM-DD.",
               "type": "str",
               "required": True
           }
       }
    }
    """

    # OpenAI tools look like this
    """
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        },
    }
    """
    cohere_tool = {
        "name": openai_tool["function"]["name"],
        "description": openai_tool["function"]["description"],
        "parameter_definitions": {},
    }

    for param_name, param_def in openai_tool["function"]["parameters"][
        "properties"
    ].items():
        required_params = (
            openai_tool.get("function", {}).get("parameters", {}).get("required", [])
        )
        cohere_param_def = {
            "description": param_def.get("description", ""),
            "type": param_def.get("type", ""),
            "required": param_name in required_params,
        }
        cohere_tool["parameter_definitions"][param_name] = cohere_param_def

    return cohere_tool


def construct_cohere_tool(tools=None):
    if tools is None:
        tools = []
    cohere_tools = []
    for tool in tools:
        cohere_tool = translate_openai_tool_to_cohere(tool)
        cohere_tools.append(cohere_tool)
    return cohere_tools


def completion(
    model: str,
    messages: list,
    api_base: str,
    model_response: ModelResponse,
    print_verbose: Callable,
    encoding,
    api_key,
    logging_obj,
    optional_params=None,
    litellm_params=None,
    logger_fn=None,
):
    headers = validate_environment(api_key)
    completion_url = api_base
    model = model
    prompt, tool_results = cohere_message_pt(messages=messages)

    ## Load Config
    config = litellm.CohereConfig.get_config()
    for k, v in config.items():
        if (
            k not in optional_params
        ):  # completion(top_k=3) > cohere_config(top_k=3) <- allows for dynamic variables to be passed in
            optional_params[k] = v

    ## Handle Tool Calling
    if "tools" in optional_params:
        _is_function_call = True
        cohere_tools = construct_cohere_tool(tools=optional_params["tools"])
        optional_params["tools"] = cohere_tools
    if len(tool_results) > 0:
        optional_params["tool_results"] = tool_results

    data = {
        "model": model,
        "message": prompt,
        **optional_params,
    }

    ## LOGGING
    logging_obj.pre_call(
        input=prompt,
        api_key=api_key,
        additional_args={
            "complete_input_dict": data,
            "headers": headers,
            "api_base": completion_url,
        },
    )
    ## COMPLETION CALL
    response = requests.post(
        completion_url,
        headers=headers,
        data=json.dumps(data),
        stream=optional_params["stream"] if "stream" in optional_params else False,
    )
    ## error handling for cohere calls
    if response.status_code != 200:
        raise CohereError(message=response.text, status_code=response.status_code)

    if "stream" in optional_params and optional_params["stream"] == True:
        return response.iter_lines()
    else:
        ## LOGGING
        logging_obj.post_call(
            input=prompt,
            api_key=api_key,
            original_response=response.text,
            additional_args={"complete_input_dict": data},
        )
        print_verbose(f"raw model_response: {response.text}")
        ## RESPONSE OBJECT
        completion_response = response.json()
        try:
            model_response.choices[0].message.content = completion_response["text"]  # type: ignore
        except Exception as e:
            raise CohereError(message=response.text, status_code=response.status_code)

        ## Tool calling response
        cohere_tools_response = completion_response.get("tool_calls", None)
        if cohere_tools_response is not None and cohere_tools_response is not []:
            # convert cohere_tools_response to OpenAI response format
            tool_calls = []
            for tool in cohere_tools_response:
                function_name = tool.get("name", "")
                generation_id = tool.get("generation_id", "")
                parameters = tool.get("parameters", {})
                tool_call = {
                    "id": f"call_{generation_id}",
                    "type": "function",
                    "function": {
                        "name": function_name,
                        "arguments": json.dumps(parameters),
                    },
                }
                tool_calls.append(tool_call)
            _message = litellm.Message(
                tool_calls=tool_calls,
                content=None,
            )
            model_response.choices[0].message = _message  # type: ignore

        ## CALCULATING USAGE - use cohere `billed_units` for returning usage
        billed_units = completion_response.get("meta", {}).get("billed_units", {})

        prompt_tokens = billed_units.get("input_tokens", 0)
        completion_tokens = billed_units.get("output_tokens", 0)

        model_response["created"] = int(time.time())
        model_response["model"] = model
        usage = Usage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
        model_response.usage = usage
        return model_response
